
Pipe2 = {};

function Pipe2.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Millloading, specializations);
end

function Pipe2:preLoad(savegame)
    self.getAllowPipeEffects2 = Utils.overwrittenFunction(self.getAllowPipeEffects, Pipe2.getAllowPipeEffects2);
    self.getOverloadingActive2 = Utils.overwrittenFunction(self.getOverloadingActive2, Pipe2.getOverloadingActive2);
end

function Pipe2:load(savegame)
    self.getIsTurnedOnAllowed2 = Utils.overwrittenFunction(self.getIsTurnedOnAllowed, Pipe2.getIsTurnedOnAllowed2);
    self.getTurnedOnNotAllowedWarning2 = Utils.overwrittenFunction(self.getTurnedOnNotAllowedWarning, Pipe2.getTurnedOnNotAllowedWarning2);
    self.getAutomaticOverloadingState2 = Utils.overwrittenFunction(self.getAutomaticOverloadingState2, Pipe2.getAutomaticOverloadingState2);

    self.getCanTipToGround2 = Pipe2.getCanTipToGround2;
    self.setDischargeToGround = SpecializationUtil.callSpecializationsFunction("setDischargeToGround");

    self.doOverload2 = Utils.overwrittenFunction(self.doOverload2, Pipe2.doOverload2);

    self.getIsFoldAllowed2 = Utils.overwrittenFunction(self.getIsFoldAllowed, Pipe2.getIsFoldAllowed2);
    self.getIsOverloadingAllowed2 = Utils.overwrittenFunction(self.getIsOverloadingAllowed2, Pipe2.getIsOverloadingAllowed2);
    self.findValidTrailer2 = Utils.overwrittenFunction(self.findValidTrailer2, Pipe2.findValidTrailer2);

    self.getIsPipeStateChangeAllowed2 = Pipe2.getIsPipeStateChangeAllowed2;
    self.setPipeState = SpecializationUtil.callSpecializationsFunction("setPipeState");
    self.findTrailerRaycastCallback2 = Pipe2.findTrailerRaycastCallback2;
    self.updatePipe2 = Pipe2.updatePipe2;

    self.setActiveControlGroup2 = Utils.overwrittenFunction(self.setActiveControlGroup, Pipe2.setActiveControlGroup2);
    self.getNextValidControlGroupIndex2 = Utils.overwrittenFunction(self.getNextValidControlGroupIndex, Pipe2.getNextValidControlGroupIndex2);
    self.getIsControlGroupChangeAllowed2 = Utils.overwrittenFunction(self.getIsControlGroupChangeAllowed, Pipe2.getIsControlGroupChangeAllowed2);

    self.pipeFoldMinTime = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#foldMinTime"), 0);
    self.pipeFoldMaxTime = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#foldMaxTime"), 1);

    self.pipeFoldMinState = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.pipe#foldMinState"), 0);
    self.pipeFoldMaxState = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.pipe#foldMaxState"), 7);
    if self.pipeFoldMaxState > 7 then
        print("Warning: pipe#foldMaxState must have a value between 0 and 7");
    end

    self.pipeNodes = {};
    self.pipeNumStates = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.pipe#numStates"), 0);
    self.pipeCurrentState = 1;
    self.pipeTargetState = 1;
    self.pipeStateIsUnloading = {};
    self.pipeStateIsAutoAiming = {};
    local unloadingPipeStates = Utils.getVectorNFromString(getXMLString(self.xmlFile, "vehicle.pipe#unloadingStates"));
    if unloadingPipeStates ~= nil then
        for i=1, table.getn(unloadingPipeStates) do
            if unloadingPipeStates[i] ~= nil then
                self.pipeStateIsUnloading[unloadingPipeStates[i] ] = true;
            end
        end
    end
    local autoAimPipeStates = Utils.getVectorNFromString(getXMLString(self.xmlFile, "vehicle.pipe#autoAimStates"));
    if autoAimPipeStates ~= nil then
        for i=1, table.getn(autoAimPipeStates) do
            if autoAimPipeStates[i] ~= nil then
                self.pipeStateIsAutoAiming[autoAimPipeStates[i] ] = true;
            end
        end
    end
    local i = 0;
    while true do
        local key = string.format("vehicle.pipe.node(%d)", i);
        if not hasXMLProperty(self.xmlFile, key) then
            break;
        end
        local node = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#index"));
        if node ~= nil then
            local entry = {};
            entry.node = node;
            entry.autoAimXRotation = Utils.getNoNil(getXMLBool(self.xmlFile, key.."#autoAimXRotation"), false);
            entry.autoAimYRotation = Utils.getNoNil(getXMLBool(self.xmlFile, key.."#autoAimYRotation"), false);
            entry.autoAimInvertZ = Utils.getNoNil(getXMLBool(self.xmlFile, key.."#autoAimInvertZ"), false);
            entry.states = {};
            for state=1,self.pipeNumStates do
                local stateKey = key..string.format(".state%d", state);
                entry.states[state] = {};
                local x,y,z = Utils.getVectorFromString(getXMLString(self.xmlFile, stateKey.."#translation"));
                if x == nil or y == nil or z == nil then
                    x,y,z = getTranslation(node);
                else
                    if state == 1 then
                        setTranslation(node, x,y,z);
                    end
                end
                entry.states[state].translation = {x,y,z};
                local x,y,z = Utils.getVectorFromString(getXMLString(self.xmlFile, stateKey.."#rotation"));
                if x == nil or y == nil or z == nil then
                    x,y,z = getRotation(node);
                else
                    x,y,z = math.rad(x),math.rad(y),math.rad(z);
                    if state == 1 then
                        setRotation(node, x,y,z);
                    end
                end
                entry.states[state].rotation = {x,y,z};
            end
            local x,y,z = Utils.getVectorFromString(getXMLString(self.xmlFile, key.."#translationSpeeds"));
            if x ~= nil and y ~= nil and z ~= nil then
                x,y,z = x*0.001,y*0.001,z*0.001;
                if x ~= 0 or y ~= 0 or z ~= 0 then
                    entry.translationSpeeds = {x,y,z};
                end
            end
            local x,y,z = Utils.getVectorFromString(getXMLString(self.xmlFile, key.."#rotationSpeeds"));
            if x ~= nil and y ~= nil and z ~= nil then
                x,y,z = math.rad(x)*0.001,math.rad(y)*0.001,math.rad(z)*0.001;
                if x ~= 0 or y ~= 0 or z ~= 0 then
                    entry.rotationSpeeds = {x,y,z};
                end
            end

            local x,y,z = getTranslation(node);
            entry.curTranslation = {x,y,z};
            local x,y,z = getRotation(node);
            entry.curRotation = {x,y,z};
            table.insert(self.pipeNodes, entry);
        end
        i = i + 1;
    end

    self.pipeAnimation = getXMLString(self.xmlFile, "vehicle.pipe#animName");
    self.pipeAnimSpeedScale = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#animSpeedScale"), 1);
    if self.pipeAnimation ~= nil then
        local targetTime = 0;
        if self.pipeAnimSpeedScale < 0 then
            targetTime = 1;
        end
        self:playAnimation(self.pipeAnimation, -self.pipeAnimSpeedScale, nil, true);
        self:setAnimationTime(self.pipeAnimation, targetTime);
    end

    self.pipeRaycastNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.pipe#raycastNodeIndex"));
    self.pipeRaycastDistance = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#raycastDistance"), 7);
    self.pipeUnloadingDistance = 0;
    self.pipeParticleSystemExtraDistance = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#effectExtraDistance"), 1);
    self.pipeParticleSystemExtraDistanceOnTrailer = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#effectExtraDistanceOnTrailer"), 0);

    self.pipeDischargeInfoIndex = getXMLInt(self.xmlFile, "vehicle.pipe#dischargeInfoIndex");

    self.trailerFound = 0;
    self.pipeFoundTrailer = false;
    self.sentPipeFoundTrailer = false;
    self.trailerFoundDistance = 0;

    self.hasPipe = self.pipeRaycastNode ~= nil;


    if self.isClient then
        self.pipeParticleSystems = {};
        local i=0;
        while true do
            local key = string.format("vehicle.pipeParticleSystems.emitterShape(%d)", i);
            if not hasXMLProperty(self.xmlFile, key) then
                break
            end

            local emitterShape = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#node"));
            local particleType = getXMLString(self.xmlFile, key.."#particleType")
            if emitterShape ~= nil then
                for _,fillUnit in pairs(self.fillUnits) do
                    for fillType,_ in pairs(fillUnit.fillTypes) do
						print("filltype: "..tostring(fillType))
                        local particleSystem = MaterialUtil.getParticleSystem(FillUtil.FILLTYPE_CCMRAW, particleType)
                        if particleSystem ~= nil then
                            if self.pipeParticleSystems[fillType] == nil then
                                self.pipeParticleSystems[fillType] = {}
                            end

                            local currentPS = ParticleUtil.copyParticleSystem(self.xmlFile, key, particleSystem, emitterShape)
                            local ps = currentPS
                            local normalSpeed,tangentSpeed = ParticleUtil.getParticleSystemAverageSpeed(ps);
                            ps.speed = math.sqrt(normalSpeed*normalSpeed + tangentSpeed*tangentSpeed);

                            table.insert(self.pipeParticleSystems[fillType], currentPS)
                        end
                    end
                end
            end

            i=i+1;
        end

        self.pipeEffects = EffectManager:loadEffect(self.xmlFile, "vehicle.pipeEffect", self.components, self);
        self.pipeTurnedOnRotationNodes = Utils.loadRotationNodes(self.xmlFile, {}, "vehicle.turnedOnRotationNodes.turnedOnRotationNode", "pipe", self.components);
        self.pipeScrollers = Utils.loadScrollers(self.components, self.xmlFile, "vehicle.pipeScrollers.pipeScroller", {}, true);

        local i = 0
        while true do
            local key = string.format("vehicle.pipe.offset(%d)", i);
            if not hasXMLProperty(self.xmlFile, key) then
                break
            end

            local movingToolNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#movingToolNode"))
            if movingToolNode ~= nil and self.nodesToMovingTools[movingToolNode] ~= nil then
                if self.pipeOffsets == nil then
                    self.pipeOffsets = {}
                end

                local offset = {}
                offset.movingTool = self.nodesToMovingTools[movingToolNode]
                offset.effects = {}
                local j = 0
                while true do
                    local effectKey = string.format(key..".effect(%d)", j);
                    if not hasXMLProperty(self.xmlFile, effectKey) then
                        break
                    end

                    local effectIndex = Utils.getNoNil(getXMLInt(self.xmlFile, effectKey.."#index"), 0)
                    if self.pipeEffects[effectIndex] ~= nil and self.pipeEffects[effectIndex].setOffset ~= nil and self.pipeEffects[effectIndex].setDelays ~= nil then
                        local effect = {}
                        effect.effect = self.pipeEffects[effectIndex]
                        effect.minValue = Utils.getNoNil(getXMLFloat(self.xmlFile, effectKey.."#minValue"), 0)
                        effect.maxValue = Utils.getNoNil(getXMLFloat(self.xmlFile, effectKey.."#maxValue"), 1)
                        local minDelay = getXMLFloat(self.xmlFile, effectKey.."#minDelay")
                        local maxDelay = getXMLFloat(self.xmlFile, effectKey.."#maxDelay")
                        effect.minDelay = Utils.getNoNil(minDelay, effect.effect.startDelay)*1000
                        effect.maxDelay = Utils.getNoNil(maxDelay, effect.effect.startDelay)*1000
                        effect.adjustDelay = minDelay ~= nil and maxDelay ~= nil
                        effect.inverted = Utils.getNoNil(getXMLBool(self.xmlFile, effectKey.."#inverted"), false)
                        table.insert(offset.effects, effect)
                    end
                    j = j + 1
                end

                table.insert(self.pipeOffsets, offset)
            end
            i = i + 1
        end
    end

    local effectIndex = getXMLInt(self.xmlFile, "vehicle.pipe#delayEffectIndex")
    if effectIndex ~= nil and self.pipeEffects[effectIndex] ~= nil then
        self.delayEffect = self.pipeEffects[effectIndex]
    end


    self.pipeBaseDelay = self.overloading.delay.time;
    self.pipeUnloadingGroundStartTime = -1;
    self.pipeUnloadingStartTime = -1;
    self.pipeScrollerEndTime = -1;

    self.pipeParticleDeactivateTime = -1;
    self.pipeUnloadEffectStopTime = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#unloadingEffectStopTime"), 0.25) * 1000;

    self.allowDischargeToGround = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.pipe#allowDischargeToGround"), true);
    self.dischargeToGround = false;
    self.remainingFillLevelToGround = 0;
    self.limitDischargeToPipeHeight = false;

    self.isSelectable = true;

    self.pipeTurnOnAllowedStates = { Utils.getVectorFromString(getXMLString(self.xmlFile, "vehicle.pipe#turnOnAllowedStates")) };
    self.pipeTurnOnStateWarning = string.format(g_i18n:getText(Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.pipe#unfoldWarning"), "warning_firstSetPipeState")), self.typeDesc);

    self.pipeDisableManualMovement = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.pipe#disableManualMovement"), true);
    self.pipeControlGroupName = Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.pipe#controlGroup"), "pipe");

    self.pipeAutomaticOverload = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.pipe#automaticOverload"), true);
    if self.pipeAutomaticOverload then
        self.overloading.canToggleOverloading = false;
    end

    self.pipeDirtyFlag = self:getNextDirtyFlag();
    self.pipeTargetTrailerDirtyFlag = self:getNextDirtyFlag();

    self:addConflictCheckedInput(InputBinding.TOGGLE_PIPE);

    self.groundUnloadingPosition = {0,0,0};
    self.groundUnloadingPipeEffect = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.pipe#groundUnloadingPipeEffect"), 1);
end

function Pipe2:postLoad(savegame)
    if savegame ~= nil and not savegame.resetVehicles then
        local pipeState = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#pipeState"), 1);
        self:setPipeState(pipeState, true);
    end
end;

function Pipe2:delete()
    if self.isClient then
        EffectManager:deleteEffects(self.pipeEffects);
        for _,particleSystems in pairs(self.pipeParticleSystems) do
            ParticleUtil.deleteParticleSystems(particleSystems)
        end
    end
end

function Pipe2:getSaveAttributesAndNodes(nodeIdent)
    local attributes = 'pipeState="'..self.pipeCurrentState..'"';
    return attributes, nil;
end

function Pipe2:readStream(streamId, connection)
    if self.hasPipe then
        local pipeState = streamReadUIntN(streamId, 2);
        self:setPipeState(pipeState, true);

        local pipeUnloadingDistance = streamReadUIntN(streamId, 8);
        self.pipeUnloadingDistance = pipeUnloadingDistance*self.pipeRaycastDistance / 255;

        if streamReadBool(streamId) then
            self.targetTrailer = readNetworkNodeObject(streamId);
        end
    end
end

function Pipe2:writeStream(streamId, connection)
    if self.hasPipe then
        streamWriteUIntN(streamId, self.pipeTargetState, 2);
        streamWriteUIntN(streamId, Utils.clamp(math.floor(self.pipeUnloadingDistance/self.pipeRaycastDistance*255), 1, 255), 8);

        if streamWriteBool(streamId, self.targetTrailer ~= nil) then
            writeNetworkNodeObject(streamId, self.targetTrailer);
        end
    end
end

function Pipe2:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then
            self.pipeUnloadingDistance = streamReadUIntN(streamId, 8)*self.pipeRaycastDistance / 255;

            self.pipeFoundTrailer = streamReadBool(streamId);
        end

        if streamReadBool(streamId) then
            self.targetTrailer = readNetworkNodeObject(streamId);
        end
    end
end

function Pipe2:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.pipeDirtyFlag) ~= 0) then
            streamWriteUIntN(streamId, Utils.clamp(math.floor(self.pipeUnloadingDistance/self.pipeRaycastDistance*255), 1, 255), 8);

            streamWriteBool(streamId, self.pipeFoundTrailer)
        end

        if streamWriteBool(streamId, bitAND(dirtyMask, self.pipeTargetTrailerDirtyFlag) ~= 0) then
            writeNetworkNodeObject(streamId, self.targetTrailer);
            self.targetTrailerSend = self.targetTrailer;
        end
    end
end

function Pipe2:mouseEvent(posX, posY, isDown, isUp, button)
end;
function Pipe2:keyEvent(unicode, sym, modifier, isDown)
end;

function Pipe2:update(dt)
    if self:getIsActive() then
        if self.isClient and self:getIsActiveForInput(true) and not self:hasInputConflictWithSelection() then
            if InputBinding.hasEvent(InputBinding.TOGGLE_PIPE) then
                if self.hasPipe then
                    local nextState = self.pipeTargetState+1;
                    if nextState > self.pipeNumStates then
                        nextState = 1;
                    end
                    if self:getIsPipeStateChangeAllowed2(nextState) then
                        self:setPipeState(nextState);
                    elseif nextState ~= 1 and self:getIsPipeStateChangeAllowed2(1) then
                        -- also try to close the pipe if other states are not allowed
                        self:setPipeState(1);
                    end
                end
            end
            if InputBinding.hasEvent(InputBinding.TOGGLE_TIPSTATE_GROUND) then
                if self.trailerFound == 0 and self:getCanTipToGround2() then
                    self:setDischargeToGround(not self.dischargeToGround);
                end
            end
        end
    end

    if self.hasPipe then
        if self.pipeStateIsAutoAiming[self.pipeTargetState] then
            self:updatePipe2(dt, self.targetTrailer);
        else
            self:updatePipe2(dt, nil);
        end
    end

    if self.isServer then
        if self.targetTrailerSend ~= self.targetTrailer then
            self:raiseDirtyFlags(self.pipeTargetTrailerDirtyFlag);
        end
    end


    if self.isClient then
        Utils.updateRotationNodes(self, self.pipeTurnedOnRotationNodes, dt, self:getAllowDelayEffects2() or self:getAllowDelayedEffects2());
        if self.pipeScrollers ~= nil then
            if self:getAllowDelayEffects2() or self:getAllowDelayedEffects2() then
                Utils.updateScrollers(self.pipeScrollers, dt, true, true, self:getUnitLastValidFillType(self.overloading.fillUnitIndex));
            else
                Utils.updateScrollers(self.pipeScrollers, dt, false, true, self:getUnitLastValidFillType(self.overloading.fillUnitIndex));
            end
        end;
    end
end

function Pipe2:updateTick(dt)
    -- if a trailer was found, check (and adjust) control group of cylindered
    if self.isServer and self.hasPipe and self:getIsActive() then

        if self.pipeStateIsUnloading[self.pipeCurrentState] and self.trailerFound ~= 0 and self.pipeDisableManualMovement then

            if self.numControlGroups ~= nil and self.numControlGroups > 1 then

                local activeGroup = self.controlGroups[self.activeControlGroupIndex];
                if activeGroup.groupName == self.pipeControlGroupName then

                    local nextIndex = self:getNextValidControlGroupIndex2();
                    if nextIndex ~= self.activeControlGroupIndex then
                        self:setActiveControlGroup2(nextIndex);
                    end
                end
            end
        end

        self.pipeFoundTrailer = false;

        if not self:getOverloadingActive2() then
            self:stopOverloadingDelay2();
        end;
    end


    if self.isServer and (self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 or self:getUnitCapacity(self.overloading.fillUnitIndex) == 0) and self.pipeStateIsUnloading[self.pipeCurrentState] then

        if self.pipeEffects ~= nil then
            if self.pipeEffects[self.groundUnloadingPipeEffect] ~= nil then
                local controlPointY = self.pipeEffects[self.groundUnloadingPipeEffect].controlPointY;
                local distance = self.pipeEffects[self.groundUnloadingPipeEffect].distance;

                if controlPointY ~= nil and distance ~= nil and controlPointY ~= 0 then
                    local mCos = math.cos(controlPointY);
                    local mSin = math.sin(controlPointY);

                    local y = Utils.dotProduct(0, 0, distance, 0.0, mCos, -mSin);
                    local z = Utils.dotProduct(0, 0, distance, 0.0, mSin, mCos);

                    if self.pipeStateIsUnloading[self.pipeTargetState] then
                        local x, y, z = localToWorld(self.pipeEffects[self.groundUnloadingPipeEffect].node, 0, y, z);
                        self.groundUnloadingPosition = {x, y, z};
                    end;
                end;
            end;
        end;

        -- do raycast
        self.trailerFound = 0;
        self.shovelTargetFound = nil;
        self.trailerFoundDistance = 0;
        if self.pipeRaycastNode ~= nil then
            local doRaycast = self.pipeStateIsUnloading[self.pipeTargetState];

            if doRaycast then
                local x,y,z = getWorldTranslation(self.pipeRaycastNode);
                local dx,dy,dz = localDirectionToWorld(self.pipeRaycastNode, 0,-1,0);
                raycastAll(x, y, z, dx,dy,dz, "findTrailerRaycastCallback2", self.pipeRaycastDistance, self);
            end

            if self.trailerFoundDistance == 0 then
                self.trailerFound = 0;
            end;

            if self.targetTrailer ~= nil then
                if self.targetTrailer.coverAnimation ~= nil then
                    if not self.targetTrailer.isCoverOpen then
                        self.trailerFound = 0;
                    end;
                end;
            end;
        end

        if self.trailerFound ~= 0 then
            if self.dischargeToGround then
                self:setDischargeToGround(false);
            end
        else
            if self.dischargeToGround then
                if not self:getCanTipToGround2() or (self.overloading.stopOverloadIfEmpty and (self:getUnitLastValidFillType(self.overloading.fillUnitIndex) == FillUtil.FILLTYPE_UNKNOWN or self:getUnitFillLevel(self.overloading.fillUnitIndex) == 0)) then
                    self:setDischargeToGround(false);
                end
            end
        end

        if self.dischargeToGround then
            if self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 and not self:getOverloadingDelayIsActive2() then
                self:startOverloadingDelay2();
            end;

            local node = self.pipeRaycastNode;
            if self.pipeDischargeInfoIndex ~= nil then
                node = self.fillVolumeDischargeInfos[self.pipeDischargeInfoIndex].nodes[1].node;
            end;

            local speed = 10;
            local x,y,z = Utils.getIntersectionOfLinearMovementAndTerrain(node, speed);
            local x0,y0,z0 = getWorldTranslation(node);
            self.pipeUnloadingDistance = Utils.vector3Length(x0-x, y0-y, z0-z);

            if self.groundUnloadingPosition[1] ~= 0 then
                x,y,z = unpack(self.groundUnloadingPosition);
            end;

            if self:getAllowDelayedUnloading2() then
                local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
                local fillLevel = self:getUnitFillLevel(self.overloading.fillUnitIndex);

                local deltaFill = fillLevel;
                if self.lastLostFillLevel ~= nil then
                    deltaFill = math.max(fillLevel, self.lastLostFillLevel);
                end

                if self:getUnitCapacity(self.overloading.fillUnitIndex) > 0 then
                    deltaFill = math.min(deltaFill, self.overloading.capacity*dt/1000.0);
                end

                self.remainingFillLevelToGround = math.min(self.remainingFillLevelToGround + deltaFill, math.max(self.overloading.capacity*dt/1000.0, TipUtil.getMinValidLiterValue(fillType)));

                if self.remainingFillLevelToGround > 0 then
                    local infos = self.fillVolumeDischargeInfos[self.overloading.dischargeInfoIndex];
                    local dx,dy,dz;
                    if infos ~= nil then
                        dx,dy,dz = localDirectionToWorld(infos.nodes[1].node, infos.nodes[1].width,0,0);
                    else
                        dx,dy,dz = localDirectionToWorld(self.components[1].node, 0,0,0.25);
                    end

                    local sx, sy, sz = x+dx, y+dy, z+dz;
                    local ex, ey, ez = x-dx, y-dy, z-dz;

                    if self.limitDischargeToPipeHeight then
                        if infos ~= nil then
                            local x,y,z = localToWorld(infos.nodes[1].node, infos.nodes[1].width,0,0);
                            sy = y;
                            local x,y,z = localToWorld(infos.nodes[1].node, -infos.nodes[1].width,0,0);
                            ey = y;
                        else
                            local x,y,z = getWorldTranslation(self.pipeRaycastNode);
                            sy = y;
                            ey = y;
                        end
                    else
                    end

                    local dropped, lineOffset = TipUtil.tipToGroundAroundLine(self, self.remainingFillLevelToGround, fillType, sx,sy,sz, ex,ey,ez, 0.1, nil, self.tipToGroundLineOffset, self.limitDischargeToPipeHeight, self.tipOcclusionAreas);

                    self.tipToGroundLineOffset = lineOffset;

                    self.remainingFillLevelToGround = self.remainingFillLevelToGround - dropped;
                    self:setUnitFillLevel(self.overloading.fillUnitIndex, fillLevel - dropped, fillType, false, self.fillVolumeUnloadInfos[self.overloading.unloadInfoIndex]);
                    self.overloading.didOverload = dropped > 0;
                end
            end;

            if not self:getCanTipToGround2() or (self.overloading.stopOverloadIfEmpty and (self:getUnitLastValidFillType(self.overloading.fillUnitIndex) == FillUtil.FILLTYPE_UNKNOWN or self:getUnitFillLevel(self.overloading.fillUnitIndex) == 0)) then
                self:setDischargeToGround(false);
            end
        elseif not self.dischargeToGround and self.overloading.isActive then
            self.pipeUnloadingDistance = self.trailerFoundDistance;
        end
    else
        self.trailerFound = 0;
    end

    if self.isServer then
        if self.pipeUnloadingDistance ~= self.sentPipeUnloadingDistance or self.sentPipeFoundTrailer ~= self.pipeFoundTrailer then
            self:raiseDirtyFlags(self.pipeDirtyFlag);

            self.sentPipeUnloadingDistance = self.pipeUnloadingDistance;
            self.sentPipeFoundTrailer = self.pipeFoundTrailer;
        end
    end

    if self.isClient then
        self.overloading.delay.time = self.pipeBaseDelay
        if self.delayEffect ~= nil then
            self.overloading.delay.time = self.delayEffect.startDelay + (1-self.delayEffect.offset)*self.delayEffect.fadeInTime
        end
        if self:getIsActiveForInput(self) then
            if self.pipeOffsets ~= nil then
                for _, offset in pairs(self.pipeOffsets) do
                    local movingTool = offset.movingTool
                    local state = Cylindered.getMovingToolState(self, movingTool);

                    for _, effect in pairs(offset.effects) do
                        local effectState = state
                        if effect.inverted then
                            effectState = 1 - effectState
                        end
                        effect.effect:setOffset(Utils.lerp(effect.minValue, effect.maxValue, effectState))
                        if effect.adjustDelay then
                            local delay = Utils.lerp(effect.minDelay, effect.maxDelay, effectState)
                            effect.effect:setDelays(delay, delay)
                        end
                    end
                end
            end
        end

        if self:getAllowDelayEffects2() then
            self.pipeParticleDeactivateTime = g_currentMission.time + self.pipeUnloadEffectStopTime;

            local currentPipeParticleSystems = self.pipeParticleSystems[self:getUnitLastValidFillType(self.overloading.fillUnitIndex)];
            if currentPipeParticleSystems ~= self.currentPipeParticleSystems then
                if self.currentPipeParticleSystems ~= nil then
                    for _, ps in pairs(self.currentPipeParticleSystems) do
                        ParticleUtil.setEmittingState(ps, false)
                    end
                end
                self.currentPipeParticleSystems = currentPipeParticleSystems
                if self.currentPipeParticleSystems ~= nil then
                    for _, ps in pairs(self.currentPipeParticleSystems) do
                        ParticleUtil.setEmittingState(ps, true)
                    end
                end
            end

            if self.currentPipeParticleSystems ~= nil then
                for _, currentPS in pairs(self.currentPipeParticleSystems) do
                    local ps = currentPS
                    if not ps.forceFullLifespan then
                        local factor = 1+((8-self.pipeUnloadingDistance-self.pipeParticleSystemExtraDistance)/15);
                        local lifespan = ps.originalLifespan * (self.pipeUnloadingDistance+self.pipeParticleSystemExtraDistance) * factor;
                        ParticleUtil.setParticleLifespan(ps, lifespan)
                    end
                end
            end

            if self.pipeEffects ~= nil then
                EffectManager:setFillType(self.pipeEffects, FillUtil.FILLTYPE_CCMRAW)
                --EffectManager:setFillType(self.pipeEffects, self:getUnitLastValidFillType(self.overloading.fillUnitIndex))
                EffectManager:startEffects(self.pipeEffects)
            end
        end

        if self.pipeEffects ~= nil then
            for _, effect in pairs(self.pipeEffects) do
                if effect.setUnloadingDistance ~= nil then
                    effect:setUnloadingDistance(self.pipeUnloadingDistance+self.pipeParticleSystemExtraDistance, g_currentMission.terrainRootNode)
                end
            end;
        end

        if self.pipeParticleDeactivateTime >= 0 and self.pipeParticleDeactivateTime <= g_currentMission.time then
            self.pipeParticleDeactivateTime = -1;
            if self.currentPipeParticleSystems ~= nil then
                for _, ps in pairs(self.currentPipeParticleSystems) do
                    ParticleUtil.setEmittingState(ps, false)
                end
                self.currentPipeParticleSystems = nil;
            end

            if self.pipeEffects ~= nil then
                EffectManager:stopEffects(self.pipeEffects);
            end
        end
    end
end

function Pipe2:draw()
    if self.isClient and self:getIsActiveForInput(true) and not self:hasInputConflictWithSelection() then
        if self.hasPipe then
            if self.pipeNumStates == 2 then
                if self.pipeTargetState == 2 then
                    if self:getIsPipeStateChangeAllowed2(1) then
                        g_currentMission:addHelpButtonText(g_i18n:getText("action_pipeIn"), InputBinding.TOGGLE_PIPE, nil, GS_PRIO_HIGH);
                    end
                else
                    if self:getIsPipeStateChangeAllowed2(2) then
                        g_currentMission:addHelpButtonText(g_i18n:getText("action_pipeOut"), InputBinding.TOGGLE_PIPE, nil, GS_PRIO_HIGH);
                    end
                end
            end
        end

        if not self.pipeFoundTrailer and self:getCanTipToGround2() then
            if self.dischargeToGround then
                g_currentMission:addHelpButtonText(g_i18n:getText("action_stopTipToGround"), InputBinding.TOGGLE_TIPSTATE_GROUND, nil, GS_PRIO_HIGH);
            elseif not self.overloading.didOverload then
                g_currentMission:addHelpButtonText(g_i18n:getText("action_startTipToGround"), InputBinding.TOGGLE_TIPSTATE_GROUND, nil, GS_PRIO_HIGH);
            end
        end
    end
end

function Pipe2:getIsTurnedOnAllowed2(superFunc, isTurnedOn)
    if isTurnedOn and self.hasPipe then
        if table.getn(self.pipeTurnOnAllowedStates) > 0 then
            local isAllowed = false;
            for _,pipeState in pairs(self.pipeTurnOnAllowedStates) do
                if pipeState == self.pipeCurrentState then
                    isAllowed = true;
                    break;
                end
            end
            if not isAllowed then
                return false;
            end
        end
    end

    if superFunc ~= nil then
        return superFunc(self, isTurnedOn);
    end

    return true;
end

function Pipe2:getAutomaticOverloadingState2(superFunc)
    local automaticOverloadingState = false
    if superFunc ~= nil then
        if superFunc(self) then
            automaticOverloadingState = true
        end
    end

    if self.trailerFound ~= 0 or self.shovelTargetFound ~= nil then
        if self.pipeAutomaticOverload then
            if self:getIsOverloadingAllowed2() and (not (self:getUnitLastValidFillType(self.overloading.fillUnitIndex) == FillUtil.FILLTYPE_UNKNOWN or self:getUnitFillLevel(self.overloading.fillUnitIndex) == 0) or not self.overloading.stopOverloadIfEmpty) then
                automaticOverloadingState = true
            end
        end
    end

    return automaticOverloadingState
end

function Pipe2:getTurnedOnNotAllowedWarning2(superFunc)
    if self.hasPipe then
        if table.getn(self.pipeTurnOnAllowedStates) > 0 then
            local isAllowed = false;
            for _,pipeState in pairs(self.pipeTurnOnAllowedStates) do
                if pipeState == self.pipeCurrentState then
                    isAllowed = true;
                    break;
                end
            end
            if not isAllowed then
                return self.pipeTurnOnStateWarning;
            end
        end
    end

    if superFunc ~= nil then
        return superFunc(self);
    end
    return nil;
end;

function Pipe2:getAllowPipeEffects2(superFunc)
    if not (self:getAllowDelayEffects2() or self:getAllowDelayedUnloading2()) then
        return false;
    end;

    if superFunc ~= nil then
        return superFunc(self);
    end
    return true;
end;

function Pipe2:getCanTipToGround2()
    if self.pipeStateIsUnloading[self.pipeCurrentState] and TipUtil.isValid() then
        local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
        if fillType ~= FillUtil.FILLTYPE_UNKNOWN then
            return TipUtil.getCanTipToGround(fillType);
        else
            return true;
        end
    else
        return false;
    end
end

function Pipe2:setDischargeToGround(state)
    SetDischargeToGroundEvent.sendEvent(self, state, noEventSend)
    self.dischargeToGround = state;
    if not state then
       self.pipeUnloadingGroundStartTime = -1;
    end;
end

function Pipe2:getOverloadingActive2(superFunc)
    if self.dischargeToGround then
        return true;
    end;

    if superFunc ~= nil then
        return superFunc(self);
    end
    return false;
end

function Pipe2:doOverload2(superFunc, targetTrailer, fillType, deltaFill)
millfillType=FillUtil.FILLTYPE_CCMRAW;
    if self.pipeRaycastNode == nil then
        return superFunc(self, targetTrailer, fillType, deltaFill);
    end


    local trailer = g_currentMission.nodeToVehicle[self.trailerFound];
    if targetTrailer == trailer and targetTrailer:getAllowFillFromAir() and deltaFill > 0 then
        targetTrailer:resetFillLevelIfNeeded(millfillType);

        local trailerFillLevel = targetTrailer:getFillLevel(millfillType);
        local trailerCapacity = targetTrailer:getCapacity(millfillType);

        deltaFill = math.min(deltaFill, trailerCapacity - trailerFillLevel);
        if deltaFill > 0 then
            targetTrailer:setFillLevel(trailerFillLevel+(deltaFill*1.2), millfillType, false, self.fillVolumeDischargeInfos[self.overloading.dischargeInfoIndex]);
            local newFillLevel = targetTrailer:getFillLevel(millfillType);
            local diff = (newFillLevel - trailerFillLevel)/1.2;
            self:setUnitFillLevel(self.overloading.fillUnitIndex, self:getUnitFillLevel(self.overloading.fillUnitIndex)-diff, fillType, false, self.fillVolumeUnloadInfos[self.overloading.unloadInfoIndex]);
            return true;
        else
            return false;
        end
    else
        return false;
    end
end

function Pipe2:getIsFoldAllowed2(superFunc, onAiTurnOn)
    if self.hasPipe and false then
        if self.overloading.didOverload or (self.dischargeToGround and self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0) then
            return false;
        end

        if self.pipeCurrentState > self.pipeFoldMaxState or self.pipeCurrentState < self.pipeFoldMinState then
            return false;
        end
    end

    if superFunc ~= nil then
        return superFunc(self, onAiTurnOn);
    end
    return true;
end

function Pipe2:getIsOverloadingAllowed2(superFunc)
    if self.hasPipe then
        if self.foldAnimTime ~= nil and (self.foldAnimTime > self.pipeFoldMaxTime or self.foldAnimTime < self.pipeFoldMinTime) then
            return false;
        end

        if not self.pipeStateIsUnloading[self.pipeCurrentState] then
            return false;
        end

        -- stop overloading if no trailer available or trailer is full or no shovel target trigger found
        if not self.dischargeToGround and self.trailerFound == 0 and self.shovelTargetFound == nil then
            return false
        end
    end

    if superFunc ~= nil then
        return superFunc(self);
    end
end

function Pipe2:findValidTrailer2(superFunc, trailers)
millfillType=FillUtil.FILLTYPE_CCMRAW;
    if self.pipeRaycastNode == nil then
        return superFunc(self, trailers);
    end

    local nearestTrailer;
    local fillType = millfillType;
   -- local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);

    -- check if a trailer is attached
    for _,implement in pairs(self.attachedImplements) do
        if implement.object ~= nil then
            if implement.object.exactFillRootNode ~= nil then
                for trailer,value in pairs(trailers) do
                    if value > 0 then
                        if entityExists(trailer.components[1].node) then
                            if trailer == implement.object and trailer:allowFillType(fillType) and trailer:getFillLevel(fillType) < trailer:getCapacity(fillType) then
                                nearestTrailer = trailer;
                                break;
                            end
                        end
                    end
                end
            end
        end
        if nearestTrailer ~= nil then
            break;
        end
    end

    -- find closest trailer to raycast node in triggers
    if nearestTrailer == nil then
	        local nearestDistance = math.huge;
	        local x0,y0,z0 = getWorldTranslation(self.pipeRaycastNode);
	
	        for trailer,value in pairs(trailers) do
	            if not entityExists(trailer.components[1].node) then
	                value = 0;
	            end
	            if value > 0 then
	                if trailer:allowFillType(fillType) and trailer.exactFillRootNode ~= nil then
	                    if trailer:getFillLevel(fillType) < trailer:getCapacity(fillType) then
	                        local x1,y1,z1 = getWorldTranslation(trailer.exactFillRootNode);
	                        local dist = Utils.vector3LengthSq(x0-x1, y0-y1, z0-z1);
	                        if dist < nearestDistance then
	                            nearestTrailer = trailer;
	                            nearestDistance = dist;
	                        end
	                    end
	                end
	            end
	        end
	    end
	
	    return nearestTrailer;
	end

	function Pipe2:findTrailerRaycastCallback2(transformId, x, y, z, distance)
	    local vehicle = g_currentMission.nodeToVehicle[transformId];
	    if vehicle ~= nil and vehicle ~= self then
	        if self.trailerFound ~= 0 and self.trailerFoundDistance ~= 0 then
	            return false;
	        end;
	
	        self.pipeFoundTrailer = true;
	
	        if vehicle.exactFillRootNode == transformId then
				local fillType = millfillType;
			 -- local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
	            if vehicle:allowFillType(fillType) and vehicle.exactFillRootNode ~= nil then
	                if vehicle:getFillLevel(fillType) < vehicle:getCapacity(fillType) then
	                    self.trailerFound = transformId;
	                end
	            end
	        else
	            self.trailerFoundDistance = distance+self.pipeParticleSystemExtraDistanceOnTrailer;
	        end
	    else
	        local object = g_currentMission:getNodeObject(transformId);
	        if object ~= nil and object.parent ~= nil and object.parent.triggerId ~= nil then
	            if object.parent:getAllowFillTypeFromTool(self:getUnitLastValidFillType(self.overloading.fillUnitIndex), TipTrigger.TOOL_TYPE_PIPE) then
	                self.shovelTargetFound = object;
	                self.trailerFoundDistance = distance+self.pipeParticleSystemExtraDistanceOnTrailer;
	            end;
	        end;
	    end
	
	    return true;
	end

	function Pipe2:getIsPipeStateChangeAllowed2(pipeState)
	    if pipeState ~= 1 and self.foldAnimTime ~= nil and (self.foldAnimTime > self.pipeFoldMaxTime or self.foldAnimTime < self.pipeFoldMinTime) then
	        return false;
	    end
	    return true;
	end

	function Pipe2:setPipeState(pipeState, noEventSend)
	    pipeState = math.min(pipeState, self.pipeNumStates);
	    if self.pipeTargetState ~= pipeState then
	        if noEventSend == nil or noEventSend == false then
	            if g_server ~= nil then
	                g_server:broadcastEvent(SetPipeStateEvent:new(self, pipeState));
	            else
	                g_client:getServerConnection():sendEvent(SetPipeStateEvent:new(self, pipeState), nil, nil, self);
	            end
	        end
	        self.pipeTargetState = pipeState;
	        self.pipeCurrentState = 0;
	        if self.pipeAnimation ~= nil then
	            if pipeState == 1 then
	                self:playAnimation(self.pipeAnimation, -self.pipeAnimSpeedScale, self:getAnimationTime(self.pipeAnimation), true);
	            else
	                self:playAnimation(self.pipeAnimation, self.pipeAnimSpeedScale, self:getAnimationTime(self.pipeAnimation), true);
	            end
	        end
	    end
	end

	function Pipe2:updatePipe2(dt, trailer)
	    if (self.pipeCurrentState ~= self.pipeTargetState or (trailer ~= nil and trailer.getAllowFillFromAir ~= nil and trailer:getAllowFillFromAir())) and self.pipeTargetState <= self.pipeNumStates then
	        local doAutoAiming = trailer ~= nil and entityExists(trailer.components[1].node) and self.pipeStateIsAutoAiming[self.pipeCurrentState];
	        local autoAimX, autoAimY, autoAimZ;
	        if doAutoAiming then
	            autoAimX, autoAimY, autoAimZ = getWorldTranslation(trailer.fillAutoAimTarget.node);
	        end
	
	        local moved = false;
	        for i=1, table.getn(self.pipeNodes) do
	            local nodeMoved = false;
	            local pipeNode = self.pipeNodes[i];
	
	            local state = pipeNode.states[self.pipeTargetState];
	            if pipeNode.translationSpeeds ~= nil then
	                for i=1, 3 do
	                    if math.abs(pipeNode.curTranslation[i] - state.translation[i]) > 0.000001 then
	                        nodeMoved = true;
	                        if pipeNode.curTranslation[i] < state.translation[i] then
	                            pipeNode.curTranslation[i] = math.min(pipeNode.curTranslation[i] + dt*pipeNode.translationSpeeds[i], state.translation[i]);
	                        else
	                            pipeNode.curTranslation[i] = math.max(pipeNode.curTranslation[i] - dt*pipeNode.translationSpeeds[i], state.translation[i]);
	                        end
	                    end
	                end
	                setTranslation(pipeNode.node, pipeNode.curTranslation[1],pipeNode.curTranslation[2],pipeNode.curTranslation[3])
	            end
	            if pipeNode.rotationSpeeds ~= nil then
	                for i=1, 3 do
	                    local targetRotation = state.rotation[i];
	                    if doAutoAiming then
	                        if pipeNode.autoAimXRotation and i == 1 then
	                            local x,y,z = getWorldTranslation(pipeNode.node);
	                            local x,y,z = worldDirectionToLocal(getParent(pipeNode.node), autoAimX-x, autoAimY-y, autoAimZ-z);
	                            targetRotation = -math.atan2(y,z);
	                            if pipeNode.autoAimInvertZ then
	                                targetRotation = targetRotation+math.pi;
	                            end
	                            targetRotation = Utils.normalizeRotationForShortestPath(targetRotation, pipeNode.curRotation[i]);
	                        elseif pipeNode.autoAimYRotation and i == 2 then
	                            local x,y,z = getWorldTranslation(pipeNode.node);
	                            local x,y,z = worldDirectionToLocal(getParent(pipeNode.node), autoAimX-x, autoAimY-y, autoAimZ-z);
	                            targetRotation = math.atan2(x,z);
	                            if pipeNode.autoAimInvertZ then
	                                targetRotation = targetRotation+math.pi;
	                            end
	                            targetRotation = Utils.normalizeRotationForShortestPath(targetRotation, pipeNode.curRotation[i]);
	                        end
	                    end
	                    if math.abs(pipeNode.curRotation[i] - targetRotation) > 0.000001 then
	                        nodeMoved = true;
	                        if pipeNode.curRotation[i] < targetRotation then
	                            pipeNode.curRotation[i] = math.min(pipeNode.curRotation[i] + dt*pipeNode.rotationSpeeds[i], targetRotation);
	                        else
	                            pipeNode.curRotation[i] = math.max(pipeNode.curRotation[i] - dt*pipeNode.rotationSpeeds[i], targetRotation);
	                        end
	                        if pipeNode.curRotation[i] > 2*math.pi then
	                            pipeNode.curRotation[i] = pipeNode.curRotation[i] - 2*math.pi;
	                        elseif pipeNode.curRotation[i] < -2*math.pi then
	                            pipeNode.curRotation[i] = pipeNode.curRotation[i] + 2*math.pi;
	                        end
	                    end
	                end
	                setRotation(pipeNode.node, pipeNode.curRotation[1],pipeNode.curRotation[2],pipeNode.curRotation[3])
	            end
	            moved = moved or nodeMoved;
	
	            if nodeMoved and self.setMovingToolDirty ~= nil then
	                self:setMovingToolDirty(pipeNode.node);
	            end
	        end
	        if table.getn(self.pipeNodes) == 0 and self.pipeAnimation ~= nil then
	            if self:getIsAnimationPlaying(self.pipeAnimation) then
	                moved = true;
	            end
	        end
	        if not moved then
	            self.pipeCurrentState = self.pipeTargetState;
	        end
	    end
	end

	function Pipe2:getNextValidControlGroupIndex2(superFunc)
	    local nextIndex = superFunc(self);
	
	    if self.pipeDisableManualMovement and self.trailerFound ~= 0 then
	        local index = self.activeControlGroupIndex;
	        for i=1,self.numControlGroups-1 do
	            index = index + 1;
	            if index > self.numControlGroups then
	                index = 1;
	            end
	
	            if self.controlGroups[index].groupName ~= self.pipeControlGroupName then
	                return index;
	            end
	        end
	        return self.activeControlGroupIndex;
	    end
	
	    return nextIndex;
	end